/* ppcrack 0.2 - decrypts PowerPacker encrypted data files with brute force
 * by Stuart Caie <kyzer@4u.net>, this software is in the Public Domain
 *
 * Webpage: http://www.kyz.uklinux.net/packers.php3#ppcrack
 *
 * For 80x86 Linux, compile with:
 * gcc -O3 -fomit-frame-pointer -fexpensive-optimizations -funroll-all-loops \
 *     -funroll-loops -finline-functions -fschedule-insns2 -mcpu=pentiumpro \
 *     -o ppcrack ppcrack.c
 *
 * Usage: ppcrack <files>
 *        ppcrack -pass <key> <files>
 *        ppcrack -key <key> <files>
 *        ppcrack -from <key> -to <key> <files>
 *
 * Examples: ppcrack -key 034b5769 AMOSdemo.AMOS.pp
 *           ppcrack -from 00000000 -to 1ffffff AMOSdemo.AMOS.pp
 *           ppcrack -pass hello *.pp
 *           ppcrack *.pp
 *
 * Anything which decrypts then decrunches to valid data is saved to disk
 * as <original filename>.<decryption key>
 *
 * As a bonus, if any file is a PowerPacker data file, but not encrypted,
 * it will be decrunched anyway, and saved as <original filename>.decrunched
 */

/* version history
 * 0.1  09-Dec-2001 - original release
 * 0.2  25-Nov-2002 - fixed off-by-one error detection error in PP_BYTE_OUT()
 *                  - allowed decrunching to end after a literal run.
 *                  - now ppValidate doesn't need the buffer decrypted.
 *                  - new command-line interface
 */

#include <stdlib.h>
#include <stdio.h>

size_t filelen;
void *loadfile(char *name) {
  void *mem = NULL; FILE *fd;
  if ((fd = fopen(name, "rb"))) {
    if ((fseek(fd, 0, SEEK_END) == 0) && (filelen = ftell(fd))
    &&  (fseek(fd, 0, SEEK_SET) == 0) && (mem = malloc(filelen))) {
      if (fread(mem, 1, filelen, fd) < filelen) { free(mem); mem = NULL; }
    }
    fclose(fd);
  }
  return mem;
}
int savefile(char *name, void *mem, size_t length) {
  FILE *fd = fopen(name, "wb");
  int ok = fd && (fwrite(mem, 1, length, fd) == length);
  if (fd) fclose(fd);
  return ok;
}

typedef unsigned char UBYTE;
typedef unsigned int ULONG;

char output_name[512];

ULONG ppCalcChecksum(UBYTE *phrase) {
  ULONG cksum = 0;
  UBYTE c, shift;

  /* for each byte in the passphrase */
  while ((c = *phrase++)) {

    /* barrel-shift the 16 bit cksum RIGHT by [c] bits */
    shift = c & 0x0F;
    if (shift) cksum = (cksum >> shift) | (cksum << (16-shift));

    /* add c to the cksum, with 16 bit wrap */
    cksum = (cksum + c) & 0xFFFF;
  }

  return cksum;
}

ULONG ppCalcPasskey(UBYTE *phrase) {
  ULONG key = 0;
  UBYTE c;

  /* for each byte in the passphrase */
  while ((c = *phrase++)) {
    /* barrel-shift 32 bit key LEFT by one bit */
    key = (key << 1) | (key >> (32-1));

    /* add c to the key, with 32 bit wrap */
    key = (key + c) & 0xFFFFFFFF;
    
    /* swap lower and upper 16 bits */
    key = (key << 16) | (key >> 16);
  }
  
  return key;
}

inline void ppDecryptCopy(UBYTE *src, UBYTE *dest, ULONG len, ULONG key) {
  UBYTE a = (key>>24) & 0xFF;
  UBYTE b = (key>>16) & 0xFF;
  UBYTE c = (key>> 8) & 0xFF;
  UBYTE d = (key    ) & 0xFF;

  len = (len + 3) >> 2;
  while (len--) {
    *dest++ = *src++ ^ a;
    *dest++ = *src++ ^ b;
    *dest++ = *src++ ^ c;
    *dest++ = *src++ ^ d;
  }
}

#define PP_READ_BITS(nbits, var) do {                          \
  bit_cnt = (nbits);                                           \
  while (bits_left < bit_cnt) {                                \
    if (buf_src < src) return 0; /* out of source bits */      \
    bit_buffer |= (*--buf_src << bits_left);                   \
    bits_left += 8;                                            \
  }                                                            \
  (var) = 0;                                                   \
  bits_left -= bit_cnt;                                        \
  while (bit_cnt--) {                                          \
    (var) = ((var) << 1) | (bit_buffer & 1);                   \
    bit_buffer >>= 1;                                          \
  }                                                            \
} while(0)

#define PP_BYTE_OUT(byte) do {                                 \
  if (out <= dest) return 0; /* output overflow */             \
  *--out = (byte);                                             \
  written++;                                                   \
} while (0)

int ppDecrunch(UBYTE *src, UBYTE *dest, UBYTE *offset_lens,
               ULONG src_len, ULONG dest_len, UBYTE skip_bits)
{
  UBYTE *buf_src, *out, *dest_end, bits_left = 0, bit_cnt;
  ULONG bit_buffer = 0, x, todo, offbits, offset, written=0;

  if (src == NULL || dest == NULL || offset_lens == NULL) return 0;

  /* set up input and output pointers */
  buf_src = src + src_len;
  out = dest_end = dest + dest_len;

  /* skip the first few bits */
  PP_READ_BITS(skip_bits, x);

  /* while there are input bits left */
  while (written < dest_len) {
    PP_READ_BITS(1, x);
    if (x == 0) {
      /* 1bit==0: literal, then match. 1bit==1: just match */
      todo = 1; do { PP_READ_BITS(2, x); todo += x; } while (x == 3);
      while (todo--) { PP_READ_BITS(8, x); PP_BYTE_OUT(x); }

      /* should we end decoding on a literal, break out of the main loop */
      if (written == dest_len) break;
    }

    /* match: read 2 bits for initial offset bitlength / match length */
    PP_READ_BITS(2, x);
    offbits = offset_lens[x];
    todo = x+2;
    if (x == 3) {
      PP_READ_BITS(1, x);
      if (x==0) offbits = 7;
      PP_READ_BITS(offbits, offset);
      do { PP_READ_BITS(3, x); todo += x; } while (x == 7);
    }
    else {
      PP_READ_BITS(offbits, offset);
    }
    if (&out[offset] >= dest_end) return 0; /* match overflow */
    while (todo--) { x = out[offset]; PP_BYTE_OUT(x); }
  }

  /* all output bytes written without error */

  /* were all source bytes used up? */
  if (buf_src > src) {
    printf("WARNING: %d source bytes were not used.\n", buf_src-src);
    printf("         decrunched file may be partially corrupt.\n");
  }
  return 1;
}                     

/* For ppValidate(), redefine PP_READ_BITS so it performs direct decryption
 * of the data stream as it's reading, so we don't have to decrypt the
 * entire stream before performing a (usually short) validation
 *
 * The key is stored in key_array[], a 4 byte array. The pointer kp is
 * predecremented to the byte of the key to be XORed with the source data.
 */
#undef PP_READ_BITS
#define PP_READ_BITS(nbits, var) do {                          \
  bit_cnt = (nbits);                                           \
  while (bits_left < bit_cnt) {                                \
    if (buf_src < src) return 0; /* out of source bits */      \
    bit_buffer |= ((*--buf_src ^ *--kp) << bits_left);         \
    if (kp == key_array) kp = &key_array[4];                   \
    bits_left += 8;                                            \
  }                                                            \
  (var) = 0;                                                   \
  bits_left -= bit_cnt;                                        \
  while (bit_cnt--) {                                          \
    (var) = ((var) << 1) | (bit_buffer & 1);                   \
    bit_buffer >>= 1;                                          \
  }                                                            \
} while(0)

/* this pretends to decrunch a data stream. If it wasn't decrypted
 * exactly right, it will access match offsets that don't exist, or
 * request match lengths that there isn't enough data for, or will
 * underrun or overrun the theoretical output buffer
 */
inline int ppValidate(UBYTE *src, UBYTE *offset_lens,
                      ULONG src_len, ULONG dest_len, UBYTE skip_bits,
                      ULONG key)
{
  UBYTE *buf_src, bits_left = 0, bit_cnt, key_array[4], *kp;
  ULONG bit_buffer = 0, x, todo, offbits, offset, written=0;
  if (src == NULL || offset_lens == NULL) return 0;

  /* set up key array and pointer */
  key_array[0] = (key >> 24) & 0xFF;
  key_array[1] = (key >> 16) & 0xFF;
  key_array[2] = (key >> 8)  & 0xFF;
  key_array[3] =  key        & 0xFF;
  kp = &key_array[4]; /* so "*--kp" will the give the last byte of the key */

  /* set up input pointer */
  buf_src = src + src_len;

  /* skip the first few bits */
  PP_READ_BITS(skip_bits, x);

  /* while there are input bits left */
  while (written < dest_len) {
    PP_READ_BITS(1, x);
    if (x == 0) {
      /* 1bit==0: literal, then match. 1bit==1: just match */
      todo = 1; do { PP_READ_BITS(2, x); todo += x; } while (x == 3);
      written += todo; if (written > dest_len) return 0;
      while (todo--) PP_READ_BITS(8, x);

      /* should we end decoding on a literal, break out of the main loop */
      if (written == dest_len) break;
    }

    /* match: read 2 bits for initial offset bitlength / match length */
    PP_READ_BITS(2, x);
    offbits = offset_lens[x];
    todo = x+2;
    if (x == 3) {
      PP_READ_BITS(1, x);
      if (x==0) offbits = 7;
      PP_READ_BITS(offbits, offset);
      do { PP_READ_BITS(3, x); todo += x; } while (x == 7);
    }
    else {
      PP_READ_BITS(offbits, offset);
    }
    if (offset >= written) return 0; /* match overflow */
    written += todo; if (written > dest_len) return 0;
  }

  /* all output bytes written without error */
  return 1;
}                     

void ppcrack(char *name, UBYTE *data, ULONG len, ULONG skey, ULONG endkey) {
  /* PP FORMAT:
   *      1 longword identifier           'PP20' or 'PX20'
   *     [1 word checksum (if 'PX20')     $ssss]
   *      1 longword efficiency           $eeeeeeee
   *      X longwords crunched file       $cccccccc,$cccccccc,...
   *      1 longword decrunch info        'decrlen' << 8 | '8 bits other info'
   */
  UBYTE *output, crypted;
  ULONG outlen;

  if (len < 16) {
    printf("%s: file is too short to be a PP file (%u bytes)\n", name, len);
    return;
  }

  if (data[0]=='P' && data[1]=='P' && data[2]=='2' && data[3]=='0') {
    if (len & 0x03) {
      printf("%s: file length is not a multiple of 4\n", name);
      return;
    }
    crypted = 0;
  }
  else if (data[0]=='P' && data[1]=='X' && data[2]=='2' && data[3]=='0') {
    if ((len-2) & 0x03) {
      printf("%s: (file length - 2) is not a multiple of 4\n", name);
      return;
    }
    crypted = 1;
    printf("%s: password checksum = %04X\n", name, (data[4]<<8) | data[5]);
  }
  else {
    printf("%s: file does not have the PP signature\n", name);
    return;
  }

  outlen = (data[len-4]<<16) | (data[len-3]<<8) | data[len-2];

  printf("%s: decrunched length = %u bytes\n", name, outlen);

  output = (UBYTE *) malloc(outlen);
  if (output == NULL) {
    printf("%s: out of memory!\n", name);
    return;
  }

  if (crypted == 0) {
    printf("%s: not encrypted, decrunching anyway\n", name);
    if (ppDecrunch(&data[8], output, &data[4], len-12, outlen, data[len-1])) {
      printf("%s: Decrunch successful!\n", name);
      sprintf(output_name, "%s.decrunched", name);
      savefile(output_name, (void *) output, outlen);
    }
    else {
      printf("%s: Decrunch failed, file is corrupt.\n", name);
    }
  }
  else {
    /* brute-force calculate the key */
    ULONG key, success=0;
    time_t time_then = 0;

    /* shortcut to halve keyspace:
     * PowerPacker alternates between two operations - literal and match.
     * The FIRST operation must be literal, as there's no data been output
     * to match yet, so the first BIT in the compressed stream must be set
     * to 0. The '8 bits other info' is actually the number of bits unused
     * in the first longword. We must ignore these.
     *
     * So we know which bit is the first one in the compressed stream, and
     * that is matched a bit in the decryption XOR key.
     *
     * We know the encrypted value of the first bit, and we know it must
     * actually be 0 when decrypted. So, if the value is 1, then that bit
     * of the decryption key must be 1, to invert that bit to a 0. If the
     * value is 0, then that bit of the decryption key must be 0, to leave
     * that bit set at 0.
     *
     * Given the knowledge of exactly one of the bits in the keys, we can
     * reject all keys that do not have the appropriate value for this bit.
     */
    ULONG drop_mask = 1 << data[len-1];
    ULONG drop_value = ( (data[len-8]<<24) | (data[len-7]<<16)
                       | (data[len-6]<<8)  |  data[len-5] ) & drop_mask;

    UBYTE *temp = (UBYTE *) malloc(len-14);
    if (temp == NULL) {
      printf("%s: out of memory!\n", name);
      return;
    }

    key = skey; key--;
    do {
      key++;
      if ((key & 0xFFF) == 0) {
        time_t time_now = time(NULL);
        /* display keys covered update once per second at most */
        if (time_now != time_then) { 
          printf("key %08X\r", key);
          fflush(stdout);
          time_then = time_now;
        }
      }

      if ((key & drop_mask) != drop_value) continue;

      if (ppValidate(&data[10], &data[6], len-14, outlen, data[len-1], key)) {
        printf("%s: key %08X success!\n", name, key);

        /* decrypt with this key */
        ppDecryptCopy(&data[10], temp, len-14, key);
	if (ppDecrunch(temp, output, &data[6], len-14, outlen, data[len-1])) {
	  success++;
	}
	else {
          printf("%s: Decrunch failed, file is corrupt.\n", name);
        }
        sprintf(output_name, "%s.%08X", name, key);
        savefile(output_name, output, outlen);
      }
    } while (key != endkey);

    free(temp);
    if (skey == endkey) {
      if (success) {
        printf("%s: the key %08X correctly decrypt the file.\n", name, skey);
      }
      else {
        printf("%s: the key %08X did not decrypt the file.\n", name, skey);
      }
    }
    else {
      printf("%s: %d successful key(s) found between %08X and %08X.\n",
             name, success, skey, endkey);
    }
  }

  free((void *) output);
}

int main(int argc, char *argv[]) {
  int i, keymode=0, passmode=0, rangemode=0;
  ULONG startkey=0x00000000, endkey=0xFFFFFFFF;
  
  if (argc < 2) {
    printf("Usage: %s [-key <key>] <file(s)>\n", argv[0]);
    printf("   or  %s [-pass <password>] <file(s)>\n", argv[0]);
    printf("   or  %s [-from <key>] [-to <key>] <file(s)>\n",argv[0]);
    printf("Example: %s -key 30a9f3e3 secret.pp\n", argv[0]);
    return 1;
  }

  /* parse any '-key', '-pass', '-from' and '-to' options */
  for (i = 1; i < argc; i++) {
    if (argv[i][0] != '-') continue;

    if (strcmp(&argv[i][1], "key") == 0) {
      if (rangemode) {
        printf("You can't mix the '-key' option with '-from' and '-to'.\n");
        return 1;
      }
      if (passmode) {
        printf("You can't mix the '-key' option with the '-pass' option.\n");
        return 1;
      }
      if (i+1 >= argc) {
        printf("Missing argument to '-key' option.\n");
        return 1;
      }
      if (sscanf(argv[i+1], "%x", &startkey) != 1) {
        printf("Bad hex number ('%s') for '-key' option.\n", argv[i+1]);
        return 1;
      }
      endkey = startkey;
      keymode = 1;
    }
    else if (strcmp(&argv[i][1], "pass") == 0) {
      if (rangemode) {
        printf("You can't mix the '-pass' option with '-from' and '-to'.\n");
        return 1;
      }
      if (keymode) {
        printf("You can't mix the '-pass' option with the '-key' option.\n");
        return 1;
      }
      if (i+1 >= argc) {
        printf("Missing argument to '-pass' option.\n");
        return 1;
      }
      startkey = endkey = ppCalcPasskey(argv[i+1]);
      printf("Password '%s' gives checksum %04X, key %08X\n", argv[i+1],
             ppCalcChecksum(argv[i+1]), startkey);
      passmode = 1;
    }
    else if (strcmp(&argv[i][1], "from") == 0) {
      if (passmode) {
        printf("You can't mix the '-from' option with the '-pass' option.\n");
        return 1;
      }
      if (keymode) {
        printf("You can't mix the '-from' option with the '-key' option.\n");
        return 1;
      }
      if (i+1 >= argc) {
        printf("Missing argument to '-from' option.\n");
        return 1;
      }
      if (sscanf(argv[i+1], "%x", &startkey) != 1) {
        printf("Bad hex number ('%s') for '-from' option.\n", argv[i+1]);
        return 1;
      }
      rangemode = 1;
    }
    else if (strcmp(&argv[i][1], "to") == 0) {
      if (passmode) {
        printf("You can't mix the '-to' option with the '-pass' option.\n");
        return 1;
      }
      if (keymode) {
        printf("You can't mix the '-to' option with the '-key' option.\n");
        return 1;
      }
      if (i+1 >= argc) {
        printf("Missing argument to '-to' option.\n");
        return 1;
      }
      if (sscanf(argv[i+1], "%x", &endkey) != 1) {
        printf("Bad hex number ('%s') for '-to' option.\n", argv[i+1]);
        return 1;
      }
      rangemode = 1;
    }
  }

  if (startkey > endkey) {
    printf("The start key (%08X) is more than the end key (%08X).\n",
           startkey, endkey);
    return 1;
  }

  if (startkey == endkey) {
    printf("Using only the key %08X to decrypt files.\n", startkey);
  }
  else {
    printf("Decrypting with keys from %08X to %08X.\n", startkey, endkey);
  }

  for (i = 1; i < argc; i++) {
    void *data;

    /* skip options */
    if (argv[i][0] == '-') {
      i++; continue;
    }

    if ((data = loadfile(argv[i])) != NULL) {
      ppcrack(argv[i], (UBYTE *) data, (int) filelen, startkey, endkey);
      free(data);
    }
    else {
      perror(argv[i]);
    }
  }
  return 0;
}
